package com.twitter.gizzard.nameserver

import java.util.TreeMap
import scala.collection.mutable
import com.twitter.logging.Logger
import com.twitter.gizzard.shards._


class NonExistentShard(message: String) extends ShardException(message: String)
class InvalidShard(message: String) extends ShardException(message: String)
class NameserverUninitialized extends ShardException("Please call reload() before operating on the NameServer")


class RoutingState(
  instantiateNode: (ShardInfo, Int, Seq[RoutingNode[Any]]) => RoutingNode[Any],
  infos: Iterable[ShardInfo],
  links: Iterable[LinkInfo],
  forwardings: Iterable[Forwarding]
) {

  val infoMap = (infos foldLeft mutable.Map[ShardId, ShardInfo]()) { (m, info) => m(info.id) = info; m }

  val linkMap = (links foldLeft mutable.Map[ShardId, List[(ShardId, Int)]]()) { (m, link) =>
    m(link.upId) = (Pair(link.downId, link.weight) :: m.getOrElse(link.upId, Nil))
    m
  }

  private def constructRoutingNode(root: ShardId, weight: Int): RoutingNode[Any] = {
    infoMap.get(root) map { rootInfo =>
      val children = linkMap.getOrElse(root, Nil) map { case (id, wt) => constructRoutingNode(id, wt) }
      instantiateNode(rootInfo, weight, children)
    } getOrElse {
      throw new InvalidShard("Expected shard '"+ root +"' to exist, but it wasn't found.")
    }
  }

  def buildForwardingTree(): scala.collection.Map[Int, TreeMap[Long, RoutingNode[Any]]] = {
    val rv = mutable.Map[Int, TreeMap[Long, RoutingNode[Any]]]()

    forwardings foreach { case Forwarding(tableId, baseId, rootShardId) =>
      val tree = rv.getOrElseUpdate(tableId, new TreeMap[Long, RoutingNode[Any]])
      tree.put(baseId, constructRoutingNode(rootShardId, 1))
    }

    rv
  }
}

class NameServer(val shard: RoutingNode[ShardManagerSource], val mappingFunction: Long => Long) {

  import ForwarderBuilder._

  private val log = Logger.get(getClass.getName)

  // XXX: inject these in later
  val shardRepository = new ShardRepository
  val shardManager    = new ShardManager(shard, shardRepository)

  @volatile private var forwardingTree: scala.collection.Map[Int, TreeMap[Long, RoutingNode[Any]]] = _

  // Forwarders

  def forwarder[T : Manifest] = {
    shardRepository.singleTableForwarder[T]
  }

  def multiTableForwarder[T : Manifest] = {
    shardRepository.multiTableForwarder[T]
  }

  def configureForwarder[T : Manifest](config: SingleForwarderBuilder[T, No, No] => SingleForwarderBuilder[T, Yes, Yes]) = {
    val forwarder = config(ForwarderBuilder.singleTable[T]).build(this)
    shardRepository.addSingleTableForwarder(forwarder)
    forwarder
  }

  def configureMultiForwarder[T : Manifest](config: MultiForwarderBuilder[T, Yes, No] => MultiForwarderBuilder[T, Yes, Yes]) = {
    val forwarder = config(ForwarderBuilder.multiTable[T]).build(this)
    shardRepository.addMultiTableForwarder(forwarder)
    forwarder
  }

  private def recreateInternalShardState() {
    val infos       = mutable.ArrayBuffer[ShardInfo]()
    val links       = mutable.ArrayBuffer[LinkInfo]()
    val forwardings = mutable.ArrayBuffer[Forwarding]()

    shardManager.currentState() foreach { state =>
      infos       ++= state.shards
      links       ++= state.links
      forwardings ++= state.forwardings
    }

    val routes = new RoutingState(
      shardRepository.instantiateNode,
      infos,
      links,
      forwardings
    )

    forwardingTree = routes.buildForwardingTree()
  }

  def reloadUpdatedForwardings() {
    log.info("Loading updated name server configuration...")
    recreateInternalShardState()
    log.info("Loading updated name server configuration is done.")
  }

  def reload() {
    log.info("Loading name server configuration...")
    shardManager.reload()
    recreateInternalShardState()
    log.info("Loading name server configuration is done.")
  }

  // XXX: removing this causes CopyJobSpec to fail.
  // This method now always falls back to the db.
  @throws(classOf[NonExistentShard])
  def findShardById[T](id: ShardId, weight: Int): RoutingNode[T] = {
    val shardInfo     = shardManager.getShard(id)
    val downwardLinks = shardManager.listDownwardLinks(id)
    val children      = downwardLinks.map(l => findShardById[T](l.downId, l.weight)).toList

    shardRepository.instantiateNode[T](shardInfo, weight, children)
  }

  // XXX: removing this causes CopyJobSpec to fail.
  // This method now always falls back to the db.
  @throws(classOf[NonExistentShard])
  def findShardById[T](id: ShardId): RoutingNode[T] = findShardById(id, 1)

  @throws(classOf[NonExistentShard])
  def findCurrentForwarding[T](tableId: Int, id: Long): RoutingNode[T] = {
    if(forwardingTree == null) throw new NameserverUninitialized

    val rv = forwardingTree.get(tableId) flatMap { treeMap =>
      treeMap.floorEntry(mappingFunction(id)) match {
        case null => None
        case item => Some(item.getValue)
      }
    } getOrElse {
      throw new NonExistentShard("No shard for address: %s %s".format(tableId, id))
    }

    // XXX: cast!
    rv.asInstanceOf[RoutingNode[T]]
  }

  @throws(classOf[NonExistentShard])
  def findForwardings[T](tableId: Int): Seq[RoutingNode[T]] = {
    import scala.collection.JavaConversions._

    if(forwardingTree == null) throw new NameserverUninitialized

    val rv = forwardingTree.get(tableId) map { bySourceIds =>
      bySourceIds.values.toSeq
    } getOrElse {
      throw new NonExistentShard("No shards for tableId: %s".format(tableId))
    }

    // XXX: cast!
    rv.asInstanceOf[Seq[RoutingNode[T]]]
  }

  def getRootShardIds(id: ShardId): Set[ShardId] = {
    val ids = shardManager.listUpwardLinks(id)
    val set: Set[ShardId] = if (ids.isEmpty) Set(id) else Set() // type needed to avoid inferring to Collection[ShardId]
    set ++ ids.flatMap((i) => getRootShardIds(i.upId).toList)
  }

  def getCommonShardId(ids: Seq[ShardId]) = {
    ids.map(getRootShardIds).reduceLeft((s1, s2) => s1.filter(s2.contains)).toSeq.headOption
  }
}
